﻿/*******************************************************************************
 * 蜂窝大战复刻版
 * -----------------------------------------------------------------------------
 * Conmajia '野比' <conmajia@gmail.com>
 * 5 JUL. 2012
 * All Rights Reserved
 * ****************************************************************************/
using System;
using System.Collections.Generic;
using System.Text;
using System.Collections;
using System.Drawing;

namespace HiveBattle
{
    public class Board
    {
        #region vars
        public DoubleBufferedGraphics gr;
        public int grid = 30;
        public int grid2;
        public int size;
        public int cx = 0;
        public int cy = 0;
        public int[,] board;
        public double[,] tmp;
        public ArrayList Players = new ArrayList();
        public int turn = 0;
        public int moves = 1;
        public int allowedmoves = 1;
        public bool picked = false;
        #endregion

        #region inits
        public Board(int size, int grid, DoubleBufferedGraphics gr)
        {
            this.gr = gr;
            this.size = size;
            this.grid = grid;
            grid2 = grid / 4;
            board = new int[size, size];
            tmp = new double[size, size];

            //Remove out of hex blocks
            for (int y = 0; y < size; y++)
            {
                for (int x = 0; x < size; x++)
                {
                    if ((x + y) < (size / 3))
                        board[x, y] = -1;
                    if ((x + size - y) < (size / 3))
                        board[x, y] = -1;
                    if ((size - x + y) < (size / 3))
                        board[x, y] = -1;
                    if ((size - x + size - y) < (size / 3))
                        board[x, y] = -1;
                }
            }

            //Create Random Holes
            Random r = new Random();
            for (int i = 0; i < ((size * size) / 30); i++)
            {
            repeat:
                int x = r.Next(size - 1);
                int y = r.Next(size - 1);
                if (board[x, y] == 0)
                    board[x, y] = -1;
                else
                    goto repeat;
            }
        }
        #endregion

        #region player
        public void FillPlayers()
        {
            Random r = new Random();

            //Fill players
            for (int t = 0; t < Players.Count; t++)
            {
                Player player = (Player)Players[t];
                for (int i = 0; i < player.Target * 6; i++)
                {
                repeat:
                    int x = r.Next(size);
                    int y = r.Next(size);
                    if (i > 0)
                    {
                        if ((board[x, y] == 0) && (IsAround(x, y, t + 1) > 0))
                            board[x, y] = t + 1;
                        else
                            goto repeat;
                    }
                    else
                    {
                        if (board[x, y] == 0)
                            board[x, y] = t + 1;
                        else
                            goto repeat;
                    }
                }

                //Fill Targets
                for (int g = 0; g < player.Target; g++)
                {
                repeat2:
                    int x = r.Next(size);
                    int y = r.Next(size);
                    if ((board[x, y] == 0) && (IsAround(x, y, 0) > 0))
                    {
                        player.x[g] = x;
                        player.y[g] = y;
                        board[x, y] = 1000;
                    }
                    else
                        goto repeat2;
                }
            }
        }
        #endregion

        #region helpers
        public int Xind(int x, int y)
        {
            x += (grid / 4);

            if (grid == 0)
                return 0;

            return (x - cx) / (grid * 3 / 2);
        }

        public int Yind(int x, int y)
        {
            x += (grid / 4);
            if (picked == false) y += grid2;

            if (grid == 0)
                return 0;

            if (((x - cx) / (grid * 3 / 2)) % 2 == 0)
                return (y - cy) / grid;
            else
                return ((y - cy) - grid / 2) / grid;
        }

        public int YSideDown(int x, int y)
        {
            return y + x % 2;
        }

        public int Visit(int x, int y, double sc, int t)
        {
            if (x < 0) return 0;
            if (y < 0) return 0;
            if (x >= size) return 0;
            if (y >= size) return 0;

            if (board[x, y] == 0)
            {
                tmp[x, y] += sc;
                return 1;
            }
            return 0;
        }

        // simple AI
        public double Cost(int player)
        {
            Player t = (Player)Players[player];
            double cost = 0;

            //Erase Tmp
            for (int y = 0; y < size; y++)
            {
                for (int x = 0; x < size; x++)
                {
                    tmp[x, y] = 0;
                }
            }

            for (int g = 0; g < t.Target; g++)
            {
                int tx = t.x[g];
                int ty = t.y[g];

                int sx;
                int sy;
                int sc = size;

                for (int l = 1; l < size; l++)
                {
                    int gaps = 0;
                    sx = tx;
                    sy = ty - l;

                    //RightDown
                    for (int h = 0; h < l; h++)
                    {
                        sy = YSideDown(sx, sy);
                        sx = sx + 1;
                        gaps += Visit(sx, sy, 1 / (double)l, player + 1);
                    }

                    //Down
                    for (int h = 0; h < l; h++)
                    {
                        sy++;
                        gaps += Visit(sx, sy, 1 / (double)l, player + 1);
                    }

                    //LeftDown
                    for (int h = 0; h < l; h++)
                    {
                        sy = YSideDown(sx, sy);
                        sx = sx - 1;
                        gaps += Visit(sx, sy, 1 / (double)l, player + 1);
                    }

                    //LeftUp
                    for (int h = 0; h < l; h++)
                    {
                        sy = YSideDown(sx, sy) - 1;
                        sx = sx - 1;
                        gaps += Visit(sx, sy, 1 / (double)l, player + 1);
                    }

                    //Up
                    for (int h = 0; h < l; h++)
                    {
                        sy--;
                        gaps += Visit(sx, sy, 1 / (double)l, player + 1);
                    }

                    //RightUp
                    for (int h = 0; h < l; h++)
                    {
                        sy = YSideDown(sx, sy) - 1;
                        sx = sx + 1;
                        gaps += Visit(sx, sy, 1 / (double)l, player + 1);
                    }

                    sc--;
                    if (gaps == 0) break;
                }
            }

            //Sum Tmp
            for (int y = 0; y < size; y++)
            {
                for (int x = 0; x < size; x++)
                {
                    cost += tmp[x, y];
                }
            }
            return cost;
        }

        int getBoardElement(int x, int y)
        {
            if ((x > -1 && x < this.size) && (y > -1 && y < this.size))
                return board[x, y];
            return -1;
        }
        public int IsAround(int x, int y, int t)
        {
            int count = 0;
            if (getBoardElement(x, y + 1) == t) count++;
            if (getBoardElement(x, y - 1) == t) count++;
            if (getBoardElement(x - 1, y) == t) count++;
            if (getBoardElement(x + 1, y) == t) count++;

            if (x % 2 == 1)
            {
                if (getBoardElement(x - 1, y + 1) == t) count++;
                if (getBoardElement(x + 1, y + 1) == t) count++;
            }
            else
            {
                if (getBoardElement(x - 1, y - 1) == t) count++;
                if (getBoardElement(x + 1, y - 1) == t) count++;
            }
            return count;
        }
        #endregion

        #region draw
        public void DrawBoard()
        {
            gr.Graphics.Clear(Color.Black);
            for (int y = 0; y < size; y++)
            {
                for (int x = 0; x < size; x++)
                {
                    if (x % 2 == 0)
                    {
                        if (board[x, y] >= 0)
                            DrawHex(cx + Xpos(x, y), cy + Ypos(x, y), Color.Black, Color.Gray);
                        if ((board[x, y] > 0) && (board[x, y] < 1000))
                            DrawHex(cx + Xpos(x, y), cy + Ypos(x, y) - grid2, Color.Black, ((Player)Players[board[x, y] - 1]).Color);
                    }
                }
                for (int x = 0; x < size; x++)
                {
                    if (x % 2 == 1)
                    {
                        if (board[x, y] >= 0)
                            DrawHex(cx + Xpos(x, y), cy + Ypos(x, y), Color.Black, Color.Gray);
                        if ((board[x, y] > 0) && (board[x, y] < 1000))
                            DrawHex(cx + Xpos(x, y), cy + Ypos(x, y) - grid2, Color.Black, ((Player)Players[board[x, y] - 1]).Color);
                    }
                }
            }
            for (int t = 0; t < Players.Count; t++)
            {
                Player player = (Player)Players[t];
                for (int g = 0; g < player.Target; g++)
                    DrawTarget(cx + Xpos(player.x[g], player.y[g]), cy + Ypos(player.x[g], player.y[g]), player.Color);
            }

            gr.Graphics.TextRenderingHint = System.Drawing.Text.TextRenderingHint.AntiAlias;
            gr.Graphics.DrawString(((Player)Players[turn]).Name + " 行动", new Font("黑体", 16, FontStyle.Regular), new SolidBrush(((Player)Players[turn]).Color), 20, 20);
        }

        public void DrawHex(int x, int y, Color c, Color fill)
        {
            Graphics g = gr.Graphics;
            Pen p = new Pen(c, 1);
            int red = fill.R - 50;
            int green = fill.G - 50;
            int blue = fill.B - 50;
            if (red < 0) red = 0;
            if (green < 0) green = 0;
            if (blue < 0) blue = 0;
            SolidBrush surfaceBrush = new SolidBrush(fill);
            SolidBrush shadowBrush = new SolidBrush(Color.FromArgb(red, green, blue));

            Point[] ptSurface = new Point[6];
            Point[] ptShadow = new Point[8];

            ptSurface[0] = new Point(x, y);
            ptSurface[1] = new Point(x + grid, y);
            ptSurface[2] = new Point(x + grid * 3 / 2, y + grid / 2);
            ptSurface[3] = new Point(x + grid, y + grid);
            ptSurface[4] = new Point(x, y + grid);
            ptSurface[5] = new Point(x - grid / 2, y + grid / 2);

            ptShadow[0] = new Point(x + grid * 3 / 2, y + grid / 2);
            ptShadow[1] = new Point(x + grid, y + grid);
            ptShadow[2] = new Point(x, y + grid);
            ptShadow[3] = new Point(x - grid / 2, y + grid / 2);
            ptShadow[4] = new Point(x - grid / 2, y + grid / 2 + grid2);
            ptShadow[5] = new Point(x, y + grid + grid2);
            ptShadow[6] = new Point(x + grid, y + grid + grid2);
            ptShadow[7] = new Point(x + grid * 3 / 2, y + grid / 2 + grid2);

            // shadow
            g.FillClosedCurve(shadowBrush, ptShadow, 0, 0.0f);
            // surface (include pieces)
            g.FillClosedCurve(surfaceBrush, ptSurface, 0, 0.0f);

            // shadow border
            g.DrawClosedCurve(p, ptShadow, 0.0f, 0);
            // piece border
            g.DrawClosedCurve(p, ptSurface, 0.0f, 0);

            // 2x shadow folded border
            g.DrawLine(p, x, y + grid, x, y + grid + grid2);
            g.DrawLine(p, x + grid, y + grid, x + grid, y + grid + grid2);
        }

        public void DrawTarget(int x, int y, Color c)
        {
            Graphics g = gr.Graphics;
            Pen p = new Pen(Color.Black, 1);
            SolidBrush br = new SolidBrush(c);

            g.FillEllipse(br, x, y + grid / 4, grid, grid / 2);
            g.DrawEllipse(p, x, y + grid / 4, grid, grid / 2);

            // draw flag
            Point[] flag = new Point[3];
            flag[0] = new Point(x + grid / 2, y - grid);
            flag[1] = new Point(x + grid / 2, y - grid / 2);
            flag[2] = new Point(x + grid * 3 / 2, y - grid / 2);

            Point pole = new Point(x + grid / 2, y + grid2);

            g.FillPolygon(br, flag);
            g.DrawLine(p, flag[0], flag[1]);
            g.DrawLine(p, flag[1], flag[2]);
            g.DrawLine(p, flag[2], flag[0]);
            g.DrawLine(p, flag[0], pole);
        }

        public int Xpos(int x, int y)
        {
            return x * grid * 3 / 2;
        }

        public int Ypos(int x, int y)
        {
            return y * grid + (x % 2) * grid / 2;
        }
        #endregion
    }
}
